// Device.h : Declaration of the CDevice

#ifndef __DEVICE_H_
#define __DEVICE_H_

#include "resource.h"       // main symbols

#include "Driver.h"


/////////////////////////////////////////////////////////////////////////////
// CDevice

class ATL_NO_VTABLE CDevice : 
   public CComObjectRootEx<CComMultiThreadModel>,
   public IPnpCallbackHardware
{
public:
   CComObject<CDriver>* m_pDriver;
   CComPtr<IWDFDevice> m_spWdfDevice;
   CComPtr<ISideShowClassExtension> m_spClassExtension;

   CDevice() : m_pDriver(NULL)
   {
   }

DECLARE_NOT_AGGREGATABLE(CDevice)
DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CDevice)
   COM_INTERFACE_ENTRY(IPnpCallbackHardware)
END_COM_MAP()

// Construction
public:
   HRESULT FinalConstruct()
   {
      HRESULT Hr;
      HR( CComObject<CDriver>::CreateInstance(&m_pDriver) );
      m_pDriver->AddRef();
      return Hr;
   }

   void FinalRelease()
   {
      if( m_pDriver != NULL ) m_pDriver->Release();
   }


// IPnpCallbackHardware
public:
   STDMETHOD(OnPrepareHardware)(IWDFDevice* pWdfDevice)
   {
      DRVTRACE(_T("CDevice::OnPrepareHardware\n"));
      VALIDATE_POINTER(pWdfDevice);
      VALIDATE_POINTER(m_pDriver);
      HRESULT Hr;
      m_spWdfDevice = pWdfDevice;
      CComPtr<IWDFNamedPropertyStore> spStore;
      HR( m_spWdfDevice->RetrieveDevicePropertyStore(NULL, WdfPropertyStoreCreateIfMissing, &spStore, NULL) );
      HR( m_pDriver->_Init(spStore) );
      if( m_spClassExtension == NULL ) {
         HR( m_spClassExtension.CoCreateInstance(CLSID_SideShowClassExtension) );
         m_spClassExtension->Initialize(m_spWdfDevice, m_pDriver);
      }
      return S_OK;
    }

   STDMETHOD(OnReleaseHardware)(IWDFDevice* pWdfDevice)
   {
      DRVTRACE(_T("CDevice::OnReleaseHardware\n"));
      VALIDATE_POINTER(pWdfDevice);
      if( m_pDriver != NULL ) m_pDriver->_Deinit();
      if( m_spClassExtension != NULL ) m_spClassExtension->Uninitialize(m_spWdfDevice);
      m_spClassExtension.Release();
      m_spWdfDevice.Release();
      return S_OK;
   }

// Operations
public:
   HRESULT _Init(IWDFDeviceInitialize* pDeviceInit)
   {
      pDeviceInit->SetLockingConstraint(WdfDeviceLevel);
      pDeviceInit->SetPowerPolicyOwnership(FALSE);
      return S_OK;
   }

   HRESULT OnProcessIoControl(IWDFIoQueue* pQueue, IWDFIoRequest* pRequest, ULONG ControlCode, SIZE_T InputBufferSizeInBytes, SIZE_T OutputBufferSizeInBytes, DWORD* pdwWritten)
   {
      DRVTRACE(_T("CDevice::OnProcessIoControl\n"));
      if( m_spClassExtension == NULL ) return E_UNEXPECTED;
      return m_spClassExtension->OnProcessIoControl(pQueue, pRequest, ControlCode, InputBufferSizeInBytes, OutputBufferSizeInBytes, pdwWritten);
   }
};


#endif //__DEVICE_H_
