// Device.h : Declaration of the CDriver

#ifndef __DRIVER_H_
#define __DRIVER_H_

#include "resource.h"       // main symbols

#include "DriverHelper.h"
#include "DisplayLink.h"


/////////////////////////////////////////////////////////////////////////////
// CDriver

class ATL_NO_VTABLE CDriver : 
   public CComObjectRootEx<CComMultiThreadModel>,
   public ISideShowDriver
{
public:
   SID* m_pUserSID;
   SIZE_T m_cbUserSID;
   CDisplayLink m_Link;
   CAppManager m_AppManager;
   CContentManager m_ContentManager;
   CSimpleArray<APPLICATION_ID> m_aAppOrder;
   CComPtr<IWDFNamedPropertyStore> m_spPropertyStore;

public:
   CDriver()
   {
      DRVTRACE(_T("CDriver::CDriver\n"));
      m_cbUserSID = SECURITY_MAX_SID_SIZE;
      m_pUserSID = (SID*) ::LocalAlloc(LMEM_FIXED | LMEM_ZEROINIT, m_cbUserSID);
   }
   ~CDriver()
   {
      DRVTRACE(_T("CDriver::~CDriver\n"));
      if( m_pUserSID != NULL ) ::LocalFree(m_pUserSID);
      m_pUserSID = NULL;
   }

DECLARE_NOT_AGGREGATABLE(CDriver)

BEGIN_COM_MAP(CDriver)
   COM_INTERFACE_ENTRY(ISideShowDriver)
END_COM_MAP()

// ISideShowDriver
public:
   STDMETHOD(OnSetCurrentUser)(const SID* pUser)
   {
      DRVTRACE(_T("CDriver::OnSetCurrentUser\n"));
      VALIDATE_POINTER(pUser);
      DRVTRACE(_T("CDriver::OnSetCurrentUser.1\n"));
      VALIDATE_POINTER(m_pUserSID);
      DRVTRACE(_T("CDriver::OnSetCurrentUser.2\n"));
      // Is it the same user?
      if( TRUE == ::EqualSid(m_pUserSID, const_cast<SID*>(pUser)) ) return S_OK;
      DRVTRACE(_T("CDriver::OnSetCurrentUser.3\n"));
      // New user logon...
      _StoreApplicationOrder();
      DRVTRACE(_T("CDriver::OnSetCurrentUser.4\n"));
      if( TRUE != ::CopySid((DWORD) m_cbUserSID, m_pUserSID, const_cast<SID*>(pUser)) ) return HRESULT_FROM_WIN32(GetLastError());
      DRVTRACE(_T("CDriver::OnSetCurrentUser.5\n"));
      m_AppManager.RemoveAll();
      DRVTRACE(_T("CDriver::OnSetCurrentUser.6\n"));
      m_ContentManager.RemoveAll();
      DRVTRACE(_T("CDriver::OnSetCurrentUser.7\n"));
      _LoadApplicationOrder();
      DRVTRACE(_T("CDriver::OnSetCurrentUser.8\n"));
      return S_OK;
   }

   STDMETHOD(OnGetCurrentUser)(SID** ppUser)
   {
      DRVTRACE(_T("CDriver::OnGetCurrentUser\n"));
      VALIDATE_POINTER(ppUser);
      VALIDATE_OUT_POINTER(ppUser);
      // We're using "console user model" so it's always the same user
//      SIZE_T cbUser = SECURITY_MAX_SID_SIZE;
      DWORD cbUser = SECURITY_MAX_SID_SIZE;
      *ppUser = (SID*) ::LocalAlloc(LMEM_FIXED, cbUser);
      if( TRUE != ::CreateWellKnownSid(WinInteractiveSid, NULL, *ppUser, &cbUser) ) return HRESULT_FROM_WIN32(GetLastError());
      return S_OK;
   }

   STDMETHOD(OnSetUserState)(const SID* /*pUser*/, const USER_STATE /*state*/)
   {
      return S_OK;
   }

   STDMETHOD(OnSetTime)(const FILETIME /*time*/)
   {
      return S_OK;
   }

   STDMETHOD(OnSetTimeZone)(const SIDESHOW_TIME_ZONE_INFORMATION* /*pTimeZoneInformation*/)
   {
      return S_OK;
   }

   STDMETHOD(OnSetShortDateFormat)(const SID* /*user*/, LPCWSTR /*wszDateFormat*/)
   {
      return S_OK;
   }

   STDMETHOD(OnSetLongDateFormat)(const SID* /*user*/, LPCWSTR /*wszDateFormat*/)
   {
      return S_OK;
   }

   STDMETHOD(OnSetShortTimeFormat)(const SID* /*user*/, LPCWSTR /*wszTimeFormat*/)
   {
      return S_OK;
   }

   STDMETHOD(OnSetLongTimeFormat)(const SID* /*user*/, LPCWSTR /*wszTimeFormat*/)
   {
      return S_OK;
   }

   STDMETHOD(OnSetTheme)(LPCWSTR /*wszTheme*/)
   {
      return S_OK;
   }

   STDMETHOD(OnGetTheme)(LPWSTR* /*pwszTheme*/)
   {
      return S_OK;
   }

   STDMETHOD(OnGetDeviceFirmwareVersion)(LPWSTR* pwszVersion)
   {
      DRVTRACE(_T("CDriver::OnGetDeviceFirmwareVersion\n"));
      return _CreateCoTaskString(L"1.0.0.0", pwszVersion);
   }

   STDMETHOD(OnGetDeviceManufacturer)(LPWSTR* pwszManufacturer)
   {
      DRVTRACE(_T("CDriver::OnGetDeviceManufacturer\n"));
      return _CreateCoTaskString(L"viksoe.dk", pwszManufacturer);
   }

   STDMETHOD(OnGetDeviceName)(LPWSTR* pwszName)
   {
      DRVTRACE(_T("CDriver::OnGetDeviceName\n"));
      return _CreateCoTaskString(L"Simple LED Display", pwszName);
   }

   STDMETHOD(OnGetDeviceEndpoints)(ENDPOINT_ID** rgEndpoints, DWORD* pcEndpoints)
   {
      DRVTRACE(_T("CDriver::OnGetDeviceEndpoints\n"));
      VALIDATE_POINTER(rgEndpoints);
      VALIDATE_POINTER(pcEndpoints);
      *rgEndpoints = (ENDPOINT_ID*) ::CoTaskMemAlloc(sizeof(ENDPOINT_ID));
      memcpy(*rgEndpoints, &SIDESHOW_ENDPOINT_SIMPLE_CONTENT_FORMAT, sizeof(SIDESHOW_ENDPOINT_SIMPLE_CONTENT_FORMAT));
      *pcEndpoints = 1;
      return S_OK;
   }

   STDMETHOD(OnGetDeviceCapabilities)(const PROPERTYKEY* /*pKey*/, PROPVARIANT* /*pvValue*/)
   {
      DRVTRACE(_T("CDriver::OnGetDeviceCapabilities\n"));
      return S_FALSE; // No support
   }

   STDMETHOD(OnGetPreEnabledApplications)(APPLICATION_ID** ppguidApps, ENDPOINT_ID** ppEndpoints, DWORD* pcApplications)
   {
      DRVTRACE(_T("CDriver::OnGetPreEnabledApplications\n"));
      VALIDATE_OUT_POINTER(ppguidApps);
      VALIDATE_OUT_POINTER(ppEndpoints);
      *pcApplications = 0;
      return S_OK;
   }

   STDMETHOD(OnGetApplicationOrder)(APPLICATION_ID** ppAppIds, DWORD* pcAppIds)
   {
      DRVTRACE(_T("CDriver::OnGetApplicationOrder\n"));
      VALIDATE_OUT_POINTER(ppAppIds);
      VALIDATE_OUT_POINTER(pcAppIds);
      CLockStaticDataInit lock;
      if( m_aAppOrder.GetSize() == 0 ) {
         for( int i = 0; i < m_AppManager.GetSize(); i++ ) {
            APPINFO Info;
            if( m_AppManager.GetInfoAt(i, Info) ) m_aAppOrder.Add(Info.AppId);
         }
      }
      int cApps = m_aAppOrder.GetSize();
      *pcAppIds = (DWORD) cApps;
      *ppAppIds = (APPLICATION_ID*) ::CoTaskMemAlloc(cApps * sizeof(APPLICATION_ID));
      for( int i = 0; i < cApps; i++ ) (*ppAppIds)[i] = m_aAppOrder[i];
      return S_OK;
   }

   STDMETHOD(OnSetApplicationOrder)(const APPLICATION_ID* pApps, const DWORD cApps)
   {
      DRVTRACE(_T("CDriver::OnSetApplicationOrder\n"));
      VALIDATE_POINTER(pApps);
      CLockStaticDataInit lock;
      m_aAppOrder.RemoveAll();
      for( DWORD i = 0; i < cApps; i++ ) {
         APPLICATION_ID AppId = pApps[i];
         m_aAppOrder.Add(AppId);
      }
      _StoreApplicationOrder();
      return S_OK;
   }

   STDMETHOD(OnAddApplication)(const SID* user,
                               REFAPPLICATION_ID rAppId,
                               REFENDPOINT_ID rEndPoint,
                               LPCWSTR wszName,
                               const DWORD cachePolicy,
                               const DWORD dwOnlineOnly,
                               const unsigned char* pbLargeIcon,
                               const DWORD cbLargeIcon,
                               const unsigned char* pbSmallIcon,
                               const DWORD cbSmallIcon,
                               const unsigned char* pbMiniIcon,
                               const DWORD cbMiniIcon)
   {
      DRVTRACE(_T("CDriver::OnAddApplication  iid=%s '%ls'\n"), DbgGetIID(rAppId), wszName);
      VALIDATE_POINTER(user);
      VALIDATE_POINTER(wszName);
      CLockStaticDataInit lock;
      APPINFO Info;
      Info.user = *user;
      Info.AppId = rAppId;
      Info.EndPoint = rEndPoint;
      ::lstrcpynW(Info.wszName, wszName, sizeof(Info.wszName) / sizeof(WCHAR));
      Info.cachePolicy = cachePolicy;
      Info.dwOnlineOnly = dwOnlineOnly;
      return m_AppManager.Add(Info) ? S_OK : E_FAIL;
   }

   STDMETHOD(OnRemoveApplication)(const SID* user, REFAPPLICATION_ID rAppId)
   {
      DRVTRACE(_T("CDriver::OnRemoveApplication  iid=%s\n"), rAppId);
      CLockStaticDataInit lock;
      m_ContentManager.RemoveApplication(rAppId);
      return m_AppManager.Remove(rAppId) ? S_OK : E_FAIL;
   }

   STDMETHOD(OnRemoveAllApplications)(const SID* user)
   {
      DRVTRACE(_T("CDriver::OnRemoveAllApplications\n"));
      CLockStaticDataInit lock;
      m_AppManager.RemoveAll();
      return S_OK;
   }

   STDMETHOD(OnSetNotificationsEnabled)(const SID* pUserSid, const BOOL fIsEnabled)
   {
      return S_OK;
   }

   STDMETHOD(OnAddNotification)(const SID* user, REFAPPLICATION_ID rAppId, const NOTIFICATION_ID notificationId, const FILETIME ftExpiration, LPCWSTR wszTitle, LPCWSTR wszMessage, const unsigned char* pbImage, const DWORD cbImage)
   {
      return S_OK;
   }

   STDMETHOD(OnRemoveNotification)(const SID* user, REFAPPLICATION_ID rAppId, const NOTIFICATION_ID notificationId)
   {
      return S_OK;
   }

   STDMETHOD(OnRemoveAllNotifications)(const SID* /*user*/, REFAPPLICATION_ID /*rAppId*/)
   {
      return S_OK;
   }

   STDMETHOD(OnAddContent)(const SID* user, REFAPPLICATION_ID rAppId, REFENDPOINT_ID rRendererId, const CONTENT_ID contentId, const unsigned char* pData, const DWORD cbData)
   {
      DRVTRACE(_T("CDriver::OnAddContent  '%ls' rendId=%ld contId=%ld\n"), m_AppManager.DbgPrintName(rAppId), rRendererId, contentId);
      
      // We're nok interested in the "glance text" which is on contentId = 0 - even if this
      // is in the most simple text format. Instead we will strip the SCF (Simple Content Format)
      // to a raw text format instead.
      if( contentId == 0 ) return S_OK;

      DRVTRACE(_T("New Content: \"%hs\"\n"), DbgGetDataBuffer(pData, cbData));

      int cbText = ::MultiByteToWideChar(CP_UTF8, 0, (LPCSTR) pData, cbData, NULL, 0);
      if( cbText == 0 ) return E_INVALIDARG;
      LPWSTR pwstr = (LPWSTR) malloc(cbText * sizeof(WCHAR));
	  if( pwstr == NULL ) return E_OUTOFMEMORY;
      ::MultiByteToWideChar(CP_UTF8, 0, (LPCSTR) pData, cbData, pwstr, cbText);
      pwstr[cbText - 1] = '\0';

      // Strip the tags from the SCF
/*
	  if( wcsncmp(pwstr, L"<body><menu", 11) == 0 ) {
	      free(pwstr);
		  return S_OK;
	  }
*/
      m_Link.ProcessContent(pwstr);

      int cchText = ::lstrlenW(pwstr);
      if( cchText == 0 ) { free(pwstr); return E_FAIL; }

      CLockStaticDataInit lock;
      CONTENTINFO Info;
      Info.user = *user;
      Info.AppId = rAppId;
      Info.RendererId = rRendererId;
      Info.ContentId = contentId;
      memset(Info.wszText, 0, sizeof(Info.wszText));
      memcpy(Info.wszText, pwstr, min(sizeof(Info.wszText) - 2, (cchText + 1) * sizeof(WCHAR)) );
      m_ContentManager.Add(Info);

      free(pwstr);
      return S_OK;
   }

   STDMETHOD(OnRemoveContent)(const SID* /*user*/, REFAPPLICATION_ID rAppId, REFENDPOINT_ID rRenderId, const CONTENT_ID contentId)
   {
      CLockStaticDataInit lock;
      m_ContentManager.Remove(rAppId, rRenderId, contentId);
      return S_OK;
   }

   STDMETHOD(OnRemoveAllContent)(const SID* /*user*/, REFAPPLICATION_ID /*rAppId*/, REFENDPOINT_ID /*rRenderId*/)
   {
      CLockStaticDataInit lock;
      m_ContentManager.RemoveAll();
      return S_OK;
   }

   STDMETHOD(OnProcessWpdMessage)(IUnknown* pinValues, IUnknown* /*poutValues*/)
   {
      DRVTRACE(_T("CDriver::OnProcessWpdMessage\n"));
      HRESULT Hr = S_OK;
      CComQIPtr<IPortableDeviceValues> spPortableDeviceValues = pinValues;
      if( spPortableDeviceValues ) return E_NOINTERFACE;
      // Check if the incoming message is a custom command we intend to act on
      /*
      PROPERTYKEY propertyKey = { 0 };
      Hr = spPortableDeviceValues->GetGuidValue(WPD_PROPERTY_COMMON_COMMAND_CATEGORY, &propertyKey.fmtid);
      if( SUCCEEDED(Hr) ) {
         Hr = spPortableDeviceValues->GetUnsignedIntegerValue(WPD_PROPERTY_COMMON_COMMAND_ID, &propertyKey.pid);
         if( SUCCEEDED(Hr) ) {
            // TODO: Process custom commands...
         }
      }
      */
      return Hr;
   }

   STDMETHOD(OnSetLanguage)(const SID* /*user*/, LPCWSTR /*wszLang*/)
   {
      DRVTRACE(_T("CDriver::OnSetLanguage\n"));
      return S_OK;
   }

// Operations
public:
   HRESULT _Init(IWDFNamedPropertyStore* pStore)
   {
      DRVTRACE(_T("CDriver::Init\n"));

      m_spPropertyStore = pStore;

      _LoadApplicationOrder();

      if( !m_Link.IsRunning() ) {
         m_Link.Stop();
         m_Link.Init(&m_ContentManager);
         m_Link.Start();
      }

      return S_OK;
   }

   void _Deinit()
   {
      DRVTRACE(_T("CDriver::Deinit\n"));
      m_Link.Abort();
      _StoreApplicationOrder();
      DRVTRACE(_T("CDriver::Deinited\n"));
   }

   void _LoadApplicationOrder()
   {
      DRVTRACE(_T("CDriver::LoadApplicationOrder\n"));
      LPVOID pVoid = NULL;
      DWORD cbData = 0;
      HRESULT Hr = _GetStoredProperty(L"ApplicationOrder", VT_BLOB, pVoid, cbData);
      DRVTRACE(_T("CDriver::LoadApplicationOrder.1\n"));
      if( FAILED(Hr) ) return;
      APPLICATION_ID* pAppIds = (APPLICATION_ID*) pVoid;
      DRVTRACE(_T("CDriver::LoadApplicationOrder.2\n"));
      int nAppIds = cbData / sizeof(APPLICATION_ID);
      OnSetApplicationOrder(pAppIds, nAppIds);
      DRVTRACE(_T("CDriver::LoadApplicationOrder.3\n"));
      ::LocalFree(pAppIds);
      DRVTRACE(_T("CDriver::LoadApplicationOrder.4\n"));
      DRVTRACE(_T("Loaded %ld AppOrderIds\n"), nAppIds);
      DRVTRACE(_T("CDriver::LoadApplicationOrder.5\n"));
   }
   void _StoreApplicationOrder()
   {
      _SetStoredProperty(L"ApplicationOrder", VT_BLOB, m_aAppOrder.GetData(), m_aAppOrder.GetSize() * sizeof(APPLICATION_ID));
   }

   HRESULT _CreateCoTaskString(LPCWSTR pstr, LPWSTR* ppstrResult) const
   {
      ATLASSERT(pstr!=NULL);
      if( ppstrResult == NULL ) return E_POINTER;
      SIZE_T cb = (wcslen(pstr) + 1) * sizeof(WCHAR);
      *ppstrResult = (LPWSTR) ::CoTaskMemAlloc(cb);
      if( *ppstrResult == NULL ) return E_OUTOFMEMORY;
      memcpy(*ppstrResult, pstr, cb);
      return S_OK;
   }

   HRESULT _GetStoredProperty(LPCWSTR pwstrKey, VARTYPE vt, LPVOID& pData, DWORD& cbData) const
   {
      VALIDATE_POINTER(pwstrKey);
      VALIDATE_POINTER(m_spPropertyStore);

      HRESULT Hr;

      LPWSTR wszSid = NULL;
      DRVTRACE(_T("CDriver::GetStoredProperty.1\n"));
      if( !::ConvertSidToStringSid(m_pUserSID, &wszSid) ) return HRESULT_FROM_WIN32(GetLastError());
      DRVTRACE(_T("CDriver::GetStoredProperty.2\n"));
      SIZE_T cchSid = ::lstrlenW(wszSid);
      DRVTRACE(_T("CDriver::GetStoredProperty.2.1\n"));
      SIZE_T cchKey = ::lstrlenW(pwstrKey);
      DRVTRACE(_T("CDriver::GetStoredProperty.2.2\n"));
      LPWSTR wszKey = (LPWSTR) _alloca(cchSid + cchKey + 4);
      DRVTRACE(_T("CDriver::GetStoredProperty.2.3\n"));
      ::wsprintfW(wszKey, L"%s -%s", pwstrKey, wszSid);
      DRVTRACE(_T("CDriver::GetStoredProperty.2.4\n"));
//      ::LocalFree(wszSid);
      DRVTRACE(_T("CDriver::GetStoredProperty.2.5\n"));

      PROPVARIANT vValue;
      ::PropVariantInit(&vValue);
      DRVTRACE(_T("CDriver::GetStoredProperty.2.6\n"));
      Hr = m_spPropertyStore->GetNamedValue(wszKey, &vValue);
      DRVTRACE(_T("CDriver::GetStoredProperty.3\n"));
      if( SUCCEEDED(Hr) && vValue.vt != vt ) Hr = E_INVALIDARG;
      if( FAILED(Hr) ) {
         ::PropVariantClear(&vValue);
         return Hr;
      }
      if( vt == VT_I4 ) {
         cbData = sizeof(LONG);
         memcpy(pData, &vValue.lVal, cbData);
      }
      if( vt == VT_R8 ) {
         cbData = sizeof(DOUBLE);
         memcpy(pData, &vValue.dblVal, cbData);
      }
      if( vt == VT_LPWSTR ) {
         cbData = min(cbData, (DWORD)::lstrlenW(vValue.pwszVal) * sizeof(WCHAR));
         memcpy(pData, vValue.pwszVal, cbData);
      }
      if( vt == VT_BLOB ) {
         cbData = vValue.blob.cbSize;
         pData = ::LocalAlloc(LPTR, cbData);
         memcpy(pData, vValue.blob.pBlobData, cbData);
      }
      DRVTRACE(_T("CDriver::GetStoredProperty.4\n"));
      ::PropVariantClear(&vValue);

      return Hr;
   }
   HRESULT _SetStoredProperty(LPCWSTR pwstrKey, VARTYPE vt, LPCVOID pData, DWORD cbData)
   {
      VALIDATE_POINTER(pwstrKey);
      VALIDATE_POINTER(m_spPropertyStore);

      HRESULT Hr;

      LPWSTR wszSid = NULL;
      if( !::ConvertSidToStringSid(m_pUserSID, &wszSid) ) return HRESULT_FROM_WIN32(GetLastError());
      SIZE_T cchSid = ::lstrlenW(wszSid);
      SIZE_T cchKey = ::lstrlenW(pwstrKey);
      LPWSTR wszKey = (LPWSTR) _alloca(cchSid + cchKey + 4);
      ::wsprintfW(wszKey, L"%s -%s", pwstrKey, wszSid);
//	  ::LocalFree(wszSid);

      PROPVARIANT vValue;
      ::PropVariantInit(&vValue);
      switch( vt ) {
      case VT_I4:
         vValue.vt = VT_I4;
         vValue.lVal = * (LONG*) pData;
         break;
      case VT_R8:
         vValue.vt = VT_R8;
         vValue.dblVal = * (DOUBLE*) pData;
         break;
      case VT_LPWSTR:
         cbData = (::lstrlenW((LPWSTR) pData) + 1) * sizeof(WCHAR);
         vValue.vt = VT_LPWSTR;
         vValue.pwszVal = (LPWSTR) ::CoTaskMemAlloc(cbData);
         memcpy(vValue.pwszVal, pData, cbData);
         break;
      case VT_BLOB:
         vValue.vt = VT_BLOB;
         vValue.blob.pBlobData = (LPBYTE) ::CoTaskMemAlloc(cbData);
         vValue.blob.cbSize = cbData;
         memcpy(vValue.blob.pBlobData, pData, cbData);
         break;
      default:
         return E_INVALIDARG;
      }
      Hr = m_spPropertyStore->SetNamedValue(wszKey, &vValue);
      ::PropVariantClear(&vValue);
      return Hr;
   }
};


#endif //__DRIVER_H_
