// DriverEntry.h : Declaration of the CDriverEntry

#ifndef __DRIVERENTRY_H_
#define __DRIVERENTRY_H_

#include "resource.h"       // main symbols

#include "Device.h"
#include "Queue.h"


/////////////////////////////////////////////////////////////////////////////
// CDriverEntry

class ATL_NO_VTABLE CDriverEntry : 
   public CComObjectRootEx<CComMultiThreadModel>,
   public CComCoClass<CDriverEntry, &CLSID_DriverEntry>,
   public IDriverEntry
{
public:
   CDriverEntry()
   {
      DRVTRACE(_T("CDriverEntry::CDriverEntry\n"));
   }

DECLARE_NO_REGISTRY()
DECLARE_NOT_AGGREGATABLE(CDriverEntry)

BEGIN_COM_MAP(CDriverEntry)
   COM_INTERFACE_ENTRY(IDriverEntry)
END_COM_MAP()

// IDriverEntry
public:
   STDMETHOD(OnInitialize)(IWDFDriver* /*pDriver*/)
   {
      DRVTRACE(_T("CDriverEntry::OnInitialize\n"));
      return S_OK;
   }

   STDMETHOD(OnDeviceAdd)(IWDFDriver* pDriver, IWDFDeviceInitialize* pDeviceInit)
   {
      DRVTRACE(_T("CDriverEntry::OnDeviceAdd\n"));
      HRESULT Hr;
      // Create and add device
      CDevicePtr pDevice = NULL;
      HR( CComObject<CDevice>::CreateInstance(&pDevice) );
      CComPtr<IUnknown> spDevice = pDevice;
      HR( pDevice->_Init(pDeviceInit) );
      CComPtr<IWDFDevice> spWDFDevice;
      HR( pDriver->CreateDevice(pDeviceInit, spDevice, &spWDFDevice) );
      // Create a default queue
      CComObject<CQueue>* pQueue = NULL;
      HR( CComObject<CQueue>::CreateInstance(&pQueue) );
      CComPtr<IUnknown> spQueue = pQueue->GetUnknown();
      HR( pQueue->_Init(pDevice) );
      CComPtr<IWDFIoQueue> spDefaultQueue;
      HR( spWDFDevice->CreateIoQueue(spQueue, TRUE, WdfIoQueueDispatchSequential, TRUE, FALSE, &spDefaultQueue) );     
      return S_OK;
   }

   STDMETHOD_(void, OnDeinitialize)(IWDFDriver* pDriver)
   {
      DRVTRACE(_T("CDriverEntry::OnDeinitialize\n"));
   }
};


#endif //__DRIVERENTRY_H_
