/////////////////////////////////////////////////////////////////////////////
// Driver Debug Logging

#ifdef _DEBUG
//#ifdef _DEBUG__

#define DRVTRACE DriverTrace
//#define DRVTRACE ATLTRACE

inline void DriverTrace(LPCTSTR pstrText, ...)
{
   USES_CONVERSION;
   // Format message with timestamp
   TCHAR szBuffer[1124] = { 0 };
   SYSTEMTIME stNow;
   ::GetLocalTime(&stNow);
   ::wsprintf(szBuffer, _T("%04ld-%02ld-%02ld %02ld:%02ld.%02ld  "),
      (long) stNow.wYear,
      (long) stNow.wMonth,
      (long) stNow.wDay,
      (long) stNow.wHour,
      (long) stNow.wMinute,
      (long) stNow.wSecond);
   TCHAR szText[1025] = { 0 };
   va_list argList;
   va_start(argList, pstrText);
   ::wvsprintf(szText, pstrText, argList);
   va_end(argList);    
   _tcscat_s(szBuffer, szText);
   // Find log filename
   static TCHAR s_szLogFilename[MAX_PATH] = { 0 };
   if( s_szLogFilename[0] == '\0' ) {
      ::GetTempPath(MAX_PATH, s_szLogFilename);
      _tcscat_s(s_szLogFilename, _T("\\Driver.log"));
   }  
   // Append text to log file
   CLockStaticDataInit lock;
   HANDLE hFile = ::CreateFile(_T("F:\\Driver.log"), GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
   if( hFile == INVALID_HANDLE_VALUE ) return;
   DWORD dwFileSize = ::SetFilePointer(hFile, 0, NULL, FILE_END);
   if( dwFileSize == INVALID_SET_FILE_POINTER ) return;
   DWORD dwBytesWritten = 0;
   LPCSTR pstr = T2CA(szBuffer);
   ::WriteFile(hFile, pstr, (DWORD)strlen(pstr), &dwBytesWritten, NULL);
   ::CloseHandle(hFile);
   ::FlushFileBuffers(hFile);
   // Log file will have a size limit...
   if( dwFileSize > 5L * 1024L * 1024L ) ::DeleteFile(s_szLogFilename);
}

// A DEBUG function which returns the CLSID or interface name
// for the IID argument.
inline LPTSTR DbgGetIID(REFIID iid)
{
   USES_CONVERSION;
   static TCHAR szName[80] = { 0 };
   szName[0] = '\0';
   OLECHAR szGUID[64];
   ::StringFromGUID2(iid, szGUID, 64);
   // Attempt to find it in the interfaces section
   CRegKey key;
   DWORD dwType = 0;
   DWORD dwCount = sizeof(szName) - 1;
   key.Open(HKEY_CLASSES_ROOT, _T("Interface"), KEY_READ);
   if( key.Open(key, OLE2T(szGUID), KEY_READ) == S_OK ) {
      ::RegQueryValueEx(key.m_hKey, NULL, NULL, &dwType, (LPBYTE) szName, &dwCount);
      return szName;
   }
   // Attempt to find it in the clsid section
   key.Open(HKEY_CLASSES_ROOT, _T("CLSID"), KEY_READ);
   if( key.Open(key, OLE2T(szGUID), KEY_READ) == S_OK ) {
      ::RegQueryValueEx(key.m_hKey, NULL, NULL, &dwType, (LPBYTE) szName, &dwCount);
      return szName;
   }
   _tcscpy_s(szName, OLE2CT(szGUID));
   return szName;
}

inline LPSTR DbgGetDataBuffer(const unsigned char* pData, DWORD cbData)
{
   static CHAR b[8500] = { 0 };
   cbData = min(sizeof(b) - 5, cbData);
   memcpy(b, pData, cbData);
//   for( DWORD i = 0; i < cbData; i++ ) if( b[i] == '\r' || b[i] == '\n' ) b[i] = ' ';
   if( strlen(b) == sizeof(b) - 5 ) strcat_s(b, "...");
   return (LPSTR) b;
}

inline LPCWSTR DbgGetStringBuffer(LPCWSTR pData)
{
   static WCHAR b[85] = { 0 };
   int cchData = min(80, (int)::lstrlen(pData) + 1);
   memcpy(b, pData, cchData * sizeof(WCHAR));
   for( int i = 0; i < cchData; i++ ) if( b[i] == '\r' || b[i] == '\n' ) b[i] = ' ';
   if( wcslen(b) == 80 ) wcscat_s(b, L"...");
   return b;
}

#else

#define DRVTRACE __noop // void(0)
#define DbgGetIID(iid) _T("")
#define DbgGetDataBuffer(pData, cbData) NULL
#define DbgGetStringBuffer(pData) NULL

#endif  // _DEBUG

