// Queue.h : Declaration of the CQueue

#ifndef __QUEUE_H_
#define __QUEUE_H_

#include "resource.h"       // main symbols


/////////////////////////////////////////////////////////////////////////////
// CQueue

class ATL_NO_VTABLE CQueue : 
   public CComObjectRootEx<CComMultiThreadModel>,
   public IQueueCallbackDeviceIoControl,
   public IQueueCallbackCreate,
   public IQueueCallbackRead,
   public IQueueCallbackWrite,
   public IFileCallbackClose,
   public IFileCallbackCleanup
{
public:
   CQueue() : m_pDevice(NULL)
   {
   }

   void FinalRelease()
   {
      if( m_pDevice != NULL ) m_pDevice->Release();
   }

DECLARE_NOT_AGGREGATABLE(CQueue)

BEGIN_COM_MAP(CQueue)
   COM_INTERFACE_ENTRY(IQueueCallbackDeviceIoControl)
   COM_INTERFACE_ENTRY(IQueueCallbackCreate)
   COM_INTERFACE_ENTRY(IQueueCallbackRead)
   COM_INTERFACE_ENTRY(IQueueCallbackWrite)
   COM_INTERFACE_ENTRY(IFileCallbackClose)
   COM_INTERFACE_ENTRY(IFileCallbackCleanup)
END_COM_MAP()

// Attributes
public:
   CDevicePtr m_pDevice;

// IQueueCallbackCreate
public:
   STDMETHOD_(void, OnCreateFile)(IWDFIoQueue* pQueue, IWDFIoRequest* pRequest, IWDFFile* pFileObject)
   {
      DRVTRACE(_T("CQueue::OnCreateFile\n"));
      pRequest->Complete(S_OK);
   }

// IQueueCallbackCleanup
public:
   STDMETHOD_(void, OnCleanupFile)(IWDFFile* pFileObject)
   {
      DRVTRACE(_T("CQueue::OnCleanupFile\n"));
   }

// IQueueCallbackClose
public:
   STDMETHOD_(void, OnCloseFile)(IWDFFile* pFileObject)
   {
      DRVTRACE(_T("CQueue::OnCloseFile\n"));
   }

// IQueueCallbackDeviceIoControl
public:
   STDMETHOD_(void, OnDeviceIoControl)(IWDFIoQueue* pQueue, IWDFIoRequest* pRequest, ULONG ControlCode, SIZE_T InputBufferSizeInBytes, SIZE_T OutputBufferSizeInBytes)
   {
      DRVTRACE(_T("CQueue::OnDeviceIoControl\n"));
      // Handle the WPD IOCTL...
      if( IS_WPD_IOCTL(ControlCode) ) {
         // Complete the request
         DWORD dwWritten = 0;
         HRESULT Hr = m_pDevice->OnProcessIoControl(pQueue, pRequest, ControlCode, InputBufferSizeInBytes, OutputBufferSizeInBytes, &dwWritten);
         pRequest->CompleteWithInformation(Hr, dwWritten);
      }
      else
      {
         // We don't handle any other types of IOCTLs.
         DWORD dwWritten = 0;
         HRESULT Hr = HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED);
         pRequest->CompleteWithInformation(Hr, dwWritten);
      }
   }

// IQueueCallbackRead
public:
   STDMETHOD_(void, OnRead)(IWDFIoQueue* /*pQueue*/, IWDFIoRequest* pRequest, SIZE_T /*NumOfBytesToRead*/)
   {
      DRVTRACE(_T("CQueue::OnRead\n"));
      HRESULT Hr = HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED);
      pRequest->CompleteWithInformation(Hr, 0);
   }

 // IQueueCallbackWrite
public:
   STDMETHOD_(void, OnWrite)(IWDFIoQueue* /*pQueue*/,IWDFIoRequest* pRequest, SIZE_T /*NumOfBytesToWrite*/)
   {
      DRVTRACE(_T("CQueue::OnWrite\n"));
      HRESULT Hr = HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED);
      pRequest->CompleteWithInformation(Hr, 0);
   }

// Operations
public:
   HRESULT _Init(CDevicePtr pDevice)
   {
      ATLASSERT(pDevice);
      m_pDevice = pDevice;
      m_pDevice->AddRef();
      return S_OK;
   }
};


#endif //__QUEUE_H_
